<?php
/* vim: set noexpandtab tabstop=2 softtabstop=2 shiftwidth=2: */

/**
 * MXInfoSearcher - Search info for TM2/SM/QM maps from ManiaExchange
 * Created by Xymph <tm@gamers.org> based on https://api2.mania-exchange.com/
 * Derived from TMXInfoSearcher
 *
 * v2.0: Updated to use the current ManiaExchange API; added MXInfo $uid,
 *       $gbxname, $gbxlogin, $authtime, $vehicle, $ismp4, $ghostblks,
 *       $embedcnt, $embedsize, $sizewarn; removed deprecated $lbrating,
 *       $replaytyp, $ratingex; replaced paginated requests with single call
 *       at limit 100; rewrote getRecent() using regular search API call
 * v1.8: Added https support to get_file; added support for TMLagoon
 * v1.7: Fixed MXInfo constructor for PHP 8.x; updated URLs to https
 * v1.6: Added Countable interface to searcher class
 * v1.5: Added MXInfo $titlepack (TM2/SM); add support for environment matching
 * v1.4: Fixed an error checking bug
 * v1.3: Added MXInfo $maptype (TM2/SM)
 * v1.2: Updated to use MX API v2.0 and add/fix support for SM; added MXInfo
 *       $trkvalue (TM2, equals deprecated $lbrating), $unlimiter (TM2/SM),
 *       $rating/$ratingex/$ratingcnt (SM)
 * v1.1: Added URLs to downloadable replays
 * v1.0: Initial release
 */
class MXInfoSearcher implements Iterator,Countable {

	public $error;
	protected $maps = array();
	private $section;
	private $prefix;

	/**
	 * Searches MX for maps matching name, author and/or environment;
	 * or search MX for the 10 most recent maps
	 *
	 * @param String $game
	 *        MX section for 'TM2', 'SM', 'QM'
	 * @param String $name
	 *        The map name to search for (partial, case-insensitive match)
	 * @param String $author
	 *        The map author to search for (partial, case-insensitive match)
	 * @param String $env
	 *        The environment to search for (exact case-sensitive match from:
	 *        TMCanyon, TMStadium, TMValley, TMLagoon, SMStorm, ...)
	 * @param Boolean $recent
	 *        If true, ignore search parameters and just return 10 newest maps
	 *        (max. one per author)
	 * @return MXInfoSearcher
	 *        If ->valid() is false, no matching map was found;
	 *        otherwise, an iterator of MXInfo objects for a 'foreach' loop.
	 *        Returns at most 100 maps.
	 */
	public function __construct($game, $name, $author, $env, $recent) {

		$this->section = $game;
		switch ($game) {
		case 'TM2':
			$this->prefix = 'tm';
			break;
		case 'SM':
			$this->prefix = 'sm';
			break;
		case 'QM':
			$this->prefix = 'qm';
			break;
		default:
			$this->prefix = '';
			return;
		}

		$this->error = '';
		if ($recent) {
			$this->maps = $this->getRecent();
		} else {
			$this->maps = $this->getList($name, $author, $env);
		}
	}  // __construct

	// define standard Iterator functions
	public function rewind() {
		reset($this->maps);
	}
	public function current() {
		return new MXInfo($this->section, $this->prefix, current($this->maps));
	}
	public function next() {
		return new MXInfo($this->section, $this->prefix, next($this->maps));
	}
	public function key() {
		return key($this->maps);
	}
	public function valid() {
		return (current($this->maps) !== false);
	}
	// define standard Countable function
	public function count() {
		return count($this->maps);
	}

	private function getRecent() {

		// get 10 most recent maps by unique author
		$url = "https://{$this->prefix}.mania-exchange.com/mapsearch2/search?api=on&limit=100";
		$file = $this->get_file($url);
		if ($file === false) {
			$this->error = 'Connection or response error on ' . $url;
			return array();
		} elseif ($file === -1) {
			$this->error = 'Timed out while reading data from ' . $url;
			return array();
		} elseif ($file == '') {
			$this->error = 'No data returned from ' . $url;
			return array();
		}

		$mx = json_decode($file);
		if ($mx === null) {
			$this->error = 'Cannot decode recent JSON data from ' . $url;
			return array();
		}

		// check for results
		$maps = array();
		if (isset($mx->results)) {
			$authors = array();
			foreach ($mx->results as $map)
				if (count($maps) < 10 && !in_array($map->UserID, $authors)) {
					$authors[] = $map->UserID;
					$maps[] = $map;
				}
		}

		// return list of maps as array of MX objects
		return $maps;
	}  // getRecent

	private function getList($name, $author, $env) {

		// compile search URL
		$url = "https://{$this->prefix}.mania-exchange.com/mapsearch2/search?api=on&limit=100";
		if ($name != '')
			$url .= '&trackname=' . rawurlencode($name);
		if ($author != '')
			$url .= '&anyauthor=' . rawurlencode($author);
		switch ($env) {
			case 'TMCanyon':
			case 'SMStorm':
				$url .= '&environments=1';
				break;
			case 'TMStadium':
				$url .= '&environments=2';
				break;
			case 'TMValley':
				$url .= '&environments=3';
				break;
			case 'TMLagoon':
				$url .= '&environments=4';
				break;
		}

		// get max. 100 results
		$file = $this->get_file($url);
		if ($file === false) {
			$this->error = 'Connection or response error on ' . $url;
			return array();
		} elseif ($file === -1) {
			$this->error = 'Timed out while reading data from ' . $url;
			return array();
		} elseif ($file == '') {
			$this->error = 'No data returned from ' . $url;
			return array();
		}

		$mx = json_decode($file);
		if ($mx === null) {
			$this->error = 'Cannot decode searched JSON data from ' . $url;
			return array();
		}
		// check for results
		$maps = array();
		if (isset($mx->results))
			$maps = $mx->results;

		// return list of maps as array of MX objects
		return $maps;
	}  // getList

	// Simple HTTP(S) Get function with timeout
	// ok: return string || error: return false || timeout: return -1
	private function get_file($url) {

		$url = parse_url($url);
		$port = isset($url['port']) ? $url['port'] :
		        ($url['scheme'] == 'https' ? 443 : 80);
		$query = isset($url['query']) ? '?' . $url['query'] : '';

		$fp = @fsockopen(($url['scheme'] == 'https' ? 'ssl://' : '') . $url['host'],
		                 $port, $errno, $errstr, 4);
		if (!$fp)
			return false;

		fwrite($fp, 'GET ' . $url['path'] . $query . " HTTP/1.0\r\n" .
		            'Host: ' . $url['host'] . "\r\n" .
		            'Content-Type: application/json' . "\r\n" .
		            'User-Agent: MXInfoSearcher (' . PHP_OS . ")\r\n\r\n");
		stream_set_timeout($fp, 4);
		$res = '';
		$info = array();
		$info['timed_out'] = false;
		while (!feof($fp) && !$info['timed_out']) {
			$res .= fread($fp, 512);
			$info = stream_get_meta_data($fp);
		}
		fclose($fp);

		if ($info['timed_out']) {
			return -1;
		} else {
			if (substr($res, 9, 3) != '200')
				return false;
			$page = explode("\r\n\r\n", $res, 2);
			return trim($page[1]);
		}
	}  // get_file
}  // class MXInfoSearcher


class MXInfo {

	public $section, $prefix, $id,
		$name, $gbxname, $uid, $userid, $author, $gbxlogin, $authtime,
		$uploaded, $updated, $type, $maptype, $titlepack, $style, $envir, $mood,
		$dispcost, $lightmap, $vehicle, $modname, $exever, $exebld, $ismp4,
		$routes, $length, $unlimiter, $laps, $diffic, $trkvalue,
		$replayid, $replaycnt, $acomment, $awards, $comments, $ghostblks,
    $embedcnt, $embedsize, $sizewarn, $rating, $ratingcnt,
		$pageurl, $replayurl, $imageurl, $thumburl, $dloadurl;

	/**
	 * Returns map object with lots of data from MX
	 *
	 * @param String $section
	 *        MX section
	 * @param String $prefix
	 *        MX URL prefix
	 * @param Object $map
	 *        The MX map data from MXInfoSearcher
	 * @return MXInfo
	 */
	public function __construct($section, $prefix, $mx) {

		$this->section  = $section;
		$this->prefix   = $prefix;
		if ($mx) {
			if ($this->prefix == 'tm')
				$dir = 'tracks';
			else // 'sm' || 'qm'
				$dir = 'maps';

			$this->id        = $this->prefix == 'tm' ? $mx->TrackID : $mx->MapID;
			$this->name      = $mx->Name;
			$this->gbxname   = $mx->GbxMapName;
			$this->uid       = $mx->TrackUID;
			$this->userid    = $mx->UserID;
			$this->author    = $mx->Username;
			$this->gbxlogin  = $mx->AuthorLogin;
			$this->authtime  = $mx->AuthorTime;
			$this->uploaded  = $mx->UploadedAt;
			$this->updated   = $mx->UpdatedAt;
			$this->type      = $mx->TypeName;
			$this->maptype   = $mx->MapType;
			$this->titlepack = $mx->TitlePack;
			$this->style     = $mx->StyleName;
			$this->envir     = $mx->EnvironmentName;
			$this->mood      = $mx->Mood;
			$this->dispcost  = $mx->DisplayCost;
			$this->lightmap  = $mx->Lightmap;
			$this->vehicle   = $mx->VehicleName;
			$this->modname   = $mx->ModName;
			$this->exever    = $mx->ExeVersion;
			$this->exebld    = $mx->ExeBuild;
			$this->ismp4     = $mx->IsMP4;
			$this->routes    = $mx->RouteName;
			$this->length    = $mx->LengthName;
			$this->unlimiter = $mx->UnlimiterRequired;
			$this->laps      = $mx->Laps;
			$this->diffic    = $mx->DifficultyName;
			$this->trkvalue  = $mx->TrackValue;
			$this->replayid  = $mx->ReplayWRID;
			$this->replaycnt = $mx->ReplayCount;
			$this->acomment  = $mx->Comments;
			$this->awards    = $mx->AwardCount;
			$this->comments  = $mx->CommentCount;
			$this->ghostblks = $mx->HasGhostBlocks;
			$this->embedcnt  = $mx->EmbeddedObjectsCount;
			$this->embedsize = $mx->EmbeddedItemsSize;
			$this->sizewarn  = $mx->SizeWarning;
			$this->rating    = $mx->RatingVoteAverage;
			$this->ratingcnt = $mx->RatingVoteCount;

			$search = array(chr(31), '[b]', '[/b]', '[i]', '[/i]', '[u]', '[/u]', '[url]', '[/url]');
			$replace = array('<br/>', '<b>', '</b>', '<i>', '</i>', '<u>', '</u>', '<i>', '</i>');
			$this->acomment  = str_ireplace($search, $replace, $this->acomment);
			$this->acomment  = preg_replace('/\[url=.*\]/', '<i>', $this->acomment);

			$this->pageurl   = "https://{$this->prefix}.mania-exchange.com/{$dir}/view/{$this->id}";
			$this->imageurl  = $mx->HasScreenshot ? "https://{$this->prefix}.mania-exchange.com/{$dir}/{$this->id}/image/1" : '';
			$this->thumburl  = $mx->HasThumbnail ? "https://{$this->prefix}.mania-exchange.com/{$dir}/thumbnail/{$this->id}" : '';
			$this->dloadurl  = "https://{$this->prefix}.mania-exchange.com/{$dir}/download/{$this->id}";

			if ($this->prefix == 'tm' && $this->replayid > 0)
				$this->replayurl = "https://{$this->prefix}.mania-exchange.com/replays/download/{$this->replayid}";
			else
				$this->replayurl = '';
		}
	}  // __construct
}  // class MXInfo

